#if defined _hook_included
    #endinput
#endif
#define _hook_included

enum
{
    HOOK_ALLOWED,
    HOOK_BLOCKED
}

enum _:HookType
{
    HT_FIXED,
    HT_FREE
}

/**
 * Sets player access to hook.
 *
 * @param id        Client index, or 0 for all players
 * @param level     Access level, 0 to remove access
 *
 * @return          true on success, false otherwise
 * @error           If the client index is greater than 0 and is not within the range of 0 to MaxClients,
 *                  an error will be thrown.
 */
native bool:hook_set_user_access(id, level = 1);

/**
 * Returns player access to hook.
 *
 * @param id        Client index
 *
 * @return          Access level
 * @error           If the client index is not within the range of 1 to MaxClients,
 *                  an error will be thrown.
 */
native hook_get_user_access(id);

/**
 * Sets player hook type.
 *
 * @param id        Client index, or 0 for all players
 * @param type      Hook type
 *
 * @return          true on success, false otherwise
 * @error           If the client index is greater than 0 and is not within the range of 0 to MaxClients,
 *                  an error will be thrown.
 */
native bool:hook_set_user_type(id, type = HT_FIXED);

/**
 * Returns player hook type.
 *
 * @param id        Client index
 *
 * @return          Hook type
 * @error           If the client index is not within the range of 1 to MaxClients,
 *                  an error will be thrown.
 */
native hook_get_user_type(id);

/**
 * Sets player hook speed.
 *
 * @param id        Client index, or 0 for all players
 * @param speed     Hook speed
 *
 * @return          true on success, false otherwise
 * @error           If the client index is greater than 0 and is not within the range of 0 to MaxClients,
 *                  an error will be thrown.
 */
native bool:hook_set_user_speed(id, Float:flSpeed);

/**
 * Returns player hook speed.
 *
 * @param id        Client index
 *
 * @return          Hook speed
 * @error           If the client index is not within the range of 1 to MaxClients,
 *                  an error will be thrown.
 */
native Float:hook_get_user_speed(id);

/**
 * Checks whether player is using hook or not.
 *
 * @param id        Client index
 *
 * @return          true if player is using it, false otherwise
 * @error           If the client index is not within the range of 1 to MaxClients,
 *                  an error will be thrown.
 */
native bool:hook_is_using(id);

/**
 * Disables hook for player.
 *
 * @param id        Client index, or 0 for all players
 *
 * @return          true on success, false otherwise
 * @error           If the client index is greater than 0 and is not within the range of 0 to MaxClients,
 *                  an error will be thrown.
 */
native bool:hook_disable(id);

/**
 * Called when player access level is changed.
 *
 * @param id        Client index
 * @param old_level Previous level access
 * @param new_level New level access
 *
 * @noreturn
 */
forward hook_access_modified(id, old_level, new_level);

/**
 * Called before player starts to use hook.
 *
 * @param id        Client index
 *
 * @return          HOOK_ALLOWED to allow
 *                  HOOK_BLOCK to deny
 */
forward hook_on_start(id);

/**
 * Called when player finished using hook.
 *
 * @param id        Client index
 *
 * @noreturn
 */
forward hook_on_finish(id);

/**
 * Called when player is using hook.
 *
 * @param id        Client index
 *
 * @return          HOOK_ALLOWED to allow
 *                  HOOK_BLOCK to stop
 */
forward hook_on_use(id);